@@ -20,8 +20,8 @@ async function ensureBrowserNavigationAllowed(url: string): Promise<void> {
 	await vscode.commands.executeCommand(ORIGIN_NAVIGATION_COMMAND, { url });
 }
 
-async function ensureCurrentPageAllowed(toolName: string): Promise<void> {
-	const currentUrl = await vscode.commands.executeCommand<string | undefined>(BROWSER_GET_URL_COMMAND);
+async function ensureCurrentPageAllowed(toolName: string, viewId: string): Promise<void> {
+	const currentUrl = await vscode.commands.executeCommand<string | undefined>(BROWSER_GET_URL_COMMAND, viewId);
 	await vscode.commands.executeCommand(ORIGIN_PAGE_COMMAND, { toolName, url: currentUrl });
 }
 
@@ -95,7 +95,7 @@ function buildPageSnapshot(depth = 0, maxDepth = 20) {
         try {
             if (!ids) return '';
             const parts = [];
-            ids.split(/\s+/).forEach(id => {
+            ids.split(/\\s+/).forEach(id => {
                 const el = document.getElementById(id);
                 if (el) {
                     const t = (el.innerText || el.textContent || '').trim();
@@ -130,7 +130,7 @@ function buildPageSnapshot(depth = 0, maxDepth = 20) {
             const parts = [];
             while (walker.nextNode()) {
                 const text = walker.currentNode.textContent || '';
-                const clean = text.replace(/\s+/g, ' ').trim();
+                const clean = text.replace(/\\s+/g, ' ').trim();
                 if (clean) {
                     parts.push(clean);
                     if (parts.join(' ').length > 240) {
@@ -141,11 +141,11 @@ function buildPageSnapshot(depth = 0, maxDepth = 20) {
             if (parts.length) {
                 return parts.join(' ').trim().substring(0, 200);
             }
-            const fallback = (el.innerText || el.textContent || '').replace(/\s+/g, ' ').trim();
+            const fallback = (el.innerText || el.textContent || '').replace(/\\s+/g, ' ').trim();
             return fallback.substring(0, 200);
         } catch (_) {
             try {
-                const text = (el.innerText || '').replace(/\s+/g, ' ').trim();
+                const text = (el.innerText || '').replace(/\\s+/g, ' ').trim();
                 return text.substring(0, 200);
             } catch (_) {
                 return '';
@@ -497,12 +497,9 @@ function buildPageSnapshot(depth = 0, maxDepth = 20) {
 }
 `;
 
-/**
- * Execute JavaScript in the browser view
- */
-async function executeInBrowser(code: string): Promise<unknown> {
+async function executeInBrowser(code: string, viewId: string): Promise<unknown> {
 	try {
-		return await vscode.commands.executeCommand('cursor.browserView.executeJavaScript', code);
+		return await vscode.commands.executeCommand('cursor.browserView.executeJavaScript', code, viewId);
 	} catch (error) {
 		CursorIDEBrowserLogger.error('Failed to execute JavaScript in browser:', error);
 		throw error;
@@ -614,21 +611,33 @@ export interface McpToolResult {
  * Browser tool implementations using direct JavaScript execution
  */
 export const BrowserTools = {
-	async navigate(params: { url: string }): Promise<McpToolResult> {
+	async navigate(params: { url: string; viewId: string; position?: 'active' | 'side' }): Promise<McpToolResult> {
 		await ensureBrowserNavigationAllowed(params.url);
 		CursorIDEBrowserLogger.info(`Navigating to ${params.url}`);
-		await vscode.commands.executeCommand('workbench.action.openBrowserEditor', {
-			preserveFocus: true,
-			url: params.url,
-			trackingId: 'mcp_tool',
-		});
-		// Wait for page to load and return snapshot
+
+		const listResult = await vscode.commands.executeCommand<{ tabs: string[] }>('cursor.browserView.listTabs');
+		const tabs = listResult?.tabs ?? [];
+
+	if (params.viewId || tabs.length > 0) {
+		const targetViewId = params.viewId || tabs[0];
+		await vscode.commands.executeCommand('cursor.browserView.navigate', params.url, targetViewId);
 		await new Promise(resolve => setTimeout(resolve, 1000));
-		return await BrowserTools.snapshot({});
+		return await BrowserTools.snapshot({ viewId: targetViewId });
+	}
+
+	const options = params.position ? { position: params.position } : undefined;
+	const result = await vscode.commands.executeCommand<{ success: boolean; browserId: string }>('cursor.browserView.newTab', params.url, options);
+	if (!result?.browserId) {
+		return {
+			content: [{ type: 'text', text: 'Failed to create new browser tab' }]
+		};
+	}
+	await new Promise(resolve => setTimeout(resolve, 1000));
+	return await BrowserTools.snapshot({ viewId: result.browserId });
 	},
 
-	async snapshot(_params: Record<string, unknown>): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_snapshot');
+	async snapshot(params: { viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_snapshot', params.viewId);
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -643,7 +652,7 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as {
+		const result = await executeInBrowser(code, params.viewId) as {
 			success: boolean;
 			pageState: PageState;
 		};
@@ -654,8 +663,8 @@ export const BrowserTools = {
 		});
 	},
 
-	async click(params: { ref: string; element?: string; doubleClick?: boolean; button?: string; modifiers?: string[] }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_click');
+	async click(params: { ref: string; element?: string; doubleClick?: boolean; button?: string; modifiers?: string[]; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_click', params.viewId);
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -716,7 +725,7 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as {
+		const result = await executeInBrowser(code, params.viewId) as {
 			success: boolean;
 			doubleClick: boolean;
 			button: string;
@@ -734,8 +743,8 @@ export const BrowserTools = {
 		});
 	},
 
-	async type(params: { ref: string; text: string; element?: string; submit?: boolean; slowly?: boolean }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_type');
+	async type(params: { ref: string; text: string; element?: string; submit?: boolean; slowly?: boolean; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_type', params.viewId);
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -804,7 +813,7 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as {
+		const result = await executeInBrowser(code, params.viewId) as {
 			success: boolean;
 			slowly: boolean;
 			submitted: boolean;
@@ -826,8 +835,8 @@ export const BrowserTools = {
 		});
 	},
 
-	async hover(params: { ref: string; element?: string }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_hover');
+	async hover(params: { ref: string; element?: string; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_hover', params.viewId);
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -860,7 +869,7 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as {
+		const result = await executeInBrowser(code, params.viewId) as {
 			success: boolean;
 			pageState: PageState;
 		};
@@ -871,8 +880,8 @@ export const BrowserTools = {
 		});
 	},
 
-	async selectOption(params: { ref: string; values: string[]; element?: string }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_select_option');
+	async selectOption(params: { ref: string; values: string[]; element?: string; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_select_option', params.viewId);
 		const code = `
 			${BROWSER_UTILS}
 			(function() {
@@ -922,7 +931,7 @@ export const BrowserTools = {
 				};
 			})();
 		`;
-		const result = await executeInBrowser(code) as {
+		const result = await executeInBrowser(code, params.viewId) as {
 			success: boolean;
 			selectedValues: string[];
 			pageState: PageState;
@@ -942,21 +951,46 @@ export const BrowserTools = {
 		});
 	},
 
-	async pressKey(params: { key: string }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_press_key');
+	async pressKey(params: { key: string; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_press_key', params.viewId);
 		const code = `
 			(function() {
 				const key = ${JSON.stringify(params.key)};
-				const activeElement = document.activeElement || document.body;
 
-				activeElement.dispatchEvent(new KeyboardEvent('keydown', { key, bubbles: true, cancelable: true }));
-				activeElement.dispatchEvent(new KeyboardEvent('keypress', { key, bubbles: true, cancelable: true }));
-				activeElement.dispatchEvent(new KeyboardEvent('keyup', { key, bubbles: true, cancelable: true }));
+				const scrollAmounts = {
+					'PageDown': { x: 0, y: window.innerHeight * 0.8 },
+					'PageUp': { x: 0, y: -window.innerHeight * 0.8 },
+					'ArrowDown': { x: 0, y: 40 },
+					'ArrowUp': { x: 0, y: -40 },
+					'End': { x: 0, y: document.documentElement.scrollHeight - window.scrollY },
+					'Home': { x: 0, y: -window.scrollY },
+					'Space': { x: 0, y: window.innerHeight * 0.8 },
+					' ': { x: 0, y: window.innerHeight * 0.8 },
+				};
+
+				const keyCodeMap = {
+					'PageDown': 34, 'PageUp': 33, 'End': 35, 'Home': 36,
+					'ArrowLeft': 37, 'ArrowUp': 38, 'ArrowRight': 39, 'ArrowDown': 40,
+					'Enter': 13, 'Escape': 27, 'Tab': 9, 'Backspace': 8, 'Delete': 46,
+					'Space': 32, ' ': 32
+				};
+
+				if (scrollAmounts[key]) {
+					window.scrollBy({ left: scrollAmounts[key].x, top: scrollAmounts[key].y, behavior: 'auto' });
+				}
+
+				const keyCode = keyCodeMap[key] || (key.length === 1 ? key.charCodeAt(0) : 0);
+				const activeElement = document.activeElement || document.body;
+				const eventInit = { key, code: key, keyCode, which: keyCode, bubbles: true, cancelable: true };
+
+				activeElement.dispatchEvent(new KeyboardEvent('keydown', eventInit));
+				activeElement.dispatchEvent(new KeyboardEvent('keypress', eventInit));
+				activeElement.dispatchEvent(new KeyboardEvent('keyup', eventInit));
 
 				return { action: 'press_key', success: true, key };
 			})();
 		`;
-		const result = await executeInBrowser(code) as { action: string; success: boolean; key: string };
+		const result = await executeInBrowser(code, params.viewId) as { action: string; success: boolean; key: string };
 		if (!result.success) {
 			return {
 				content: [{ type: 'text', text: `Failed to press key: ${result.key}` }]
@@ -967,8 +1001,8 @@ export const BrowserTools = {
 		};
 	},
 
-	async waitFor(params: { time?: number; text?: string; textGone?: string }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_wait_for');
+	async waitFor(params: { time?: number; text?: string; textGone?: string; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_wait_for', params.viewId);
 		if (params.time !== undefined) {
 			await new Promise(resolve => setTimeout(resolve, params.time! * 1000));
 			return {
@@ -1012,7 +1046,7 @@ export const BrowserTools = {
 				});
 			})();
 		`;
-		const { success, type, text, error } = await executeInBrowser(code) as { action: string; success: boolean; type: string; text?: string; error?: string };
+		const { success, type, text, error } = await executeInBrowser(code, params.viewId) as { action: string; success: boolean; type: string; text?: string; error?: string };
 		if (!success) {
 			return {
 				content: [{ type: 'text', text: `Failed to wait for element: ${error || 'Unknown error'}` }]
@@ -1023,32 +1057,30 @@ export const BrowserTools = {
 		};
 	},
 
-	async consoleMessages(_params: Record<string, unknown>): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_console_messages');
-		// Console logs are now tracked by the main process
-		const logs = await vscode.commands.executeCommand<Array<{ type: string; message: string; timestamp: number }>>('cursor.browserView.getConsoleLogs');
+	async consoleMessages(params: { viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_console_messages', params.viewId);
+		const logs = await vscode.commands.executeCommand<Array<{ type: string; message: string; timestamp: number }>>('cursor.browserView.getConsoleLogs', params.viewId);
 		return {
 			content: [{ type: 'text', text: JSON.stringify(logs, null, 2) }]
 		};
 	},
 
-	async networkRequests(_params: Record<string, unknown>): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_network_requests');
-		// Network requests are now tracked by the main process
+	async networkRequests(params: { viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_network_requests', params.viewId);
 		const requests = await vscode.commands.executeCommand<Array<{
 			url: string;
 			method: string;
 			statusCode?: number;
 			timestamp: number;
 			resourceType?: string;
-		}>>('cursor.browserView.getNetworkRequests');
+		}>>('cursor.browserView.getNetworkRequests', params.viewId);
 		return {
 			content: [{ type: 'text', text: JSON.stringify(requests, null, 2) }]
 		};
 	},
 
-	async takeScreenshot(params: { filename?: string; type?: string; fullPage?: boolean; element?: string; ref?: string }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_take_screenshot');
+	async takeScreenshot(params: { filename?: string; type?: string; fullPage?: boolean; element?: string; ref?: string; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_take_screenshot', params.viewId);
 
 		// Ensure filename has the correct extension if provided without one
 		let processedParams = params;
@@ -1123,9 +1155,9 @@ export const BrowserTools = {
 		};
 	},
 
-	async goBack(_params: Record<string, unknown>): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_navigate_back');
-		const success = await vscode.commands.executeCommand('cursor.browserView.goBack') as boolean;
+	async goBack(params: { viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_navigate_back', params.viewId);
+		const success = await vscode.commands.executeCommand('cursor.browserView.goBack', params.viewId) as boolean;
 		if (!success) {
 			return {
 				content: [{ type: 'text', text: 'Failed to navigate back' }]
@@ -1136,9 +1168,9 @@ export const BrowserTools = {
 		};
 	},
 
-	async resize(params: { width: number; height: number }): Promise<McpToolResult> {
-		await ensureCurrentPageAllowed('browser_resize');
-		const { success } = await vscode.commands.executeCommand('cursor.browserView.resize', params) as { success: boolean };
+	async resize(params: { width: number; height: number; viewId: string }): Promise<McpToolResult> {
+		await ensureCurrentPageAllowed('browser_resize', params.viewId);
+		const { success } = await vscode.commands.executeCommand('cursor.browserView.resize', { ...params, viewId: params.viewId }) as { success: boolean };
 		if (!success) {
 			return {
 				content: [{ type: 'text', text: 'Failed to resize browser' }]
@@ -1147,6 +1179,54 @@ export const BrowserTools = {
 		return {
 			content: [{ type: 'text', text: 'Successfully resized browser' }]
 		};
+	},
+
+	async tabs(params: { action: string; index?: number; position?: 'active' | 'side' }): Promise<McpToolResult> {
+		switch (params.action) {
+			case 'list': {
+				const result = await vscode.commands.executeCommand<{ tabs: string[]; activeTab?: string; lastInteractedTab?: string }>('cursor.browserView.listTabs');
+				if (!result || !result.tabs) {
+					return { content: [{ type: 'text', text: 'Open tabs:\n(none)' }] };
+				}
+				const tabInfo = await Promise.all(result.tabs.map(async (tabId: string, idx: number) => {
+					const url = await vscode.commands.executeCommand<string>('cursor.browserView.getURL', tabId);
+					return `[${idx}] (viewId: ${tabId}) ${url || 'about:blank'}`;
+				}));
+				return { content: [{ type: 'text', text: `Open tabs:\n${tabInfo.join('\n')}` }] };
+			}
+			case 'new': {
+				const options = params.position ? { position: params.position } : undefined;
+				await vscode.commands.executeCommand<{ success: boolean; browserId: string }>('cursor.browserView.newTab', undefined, options);
+				return { content: [{ type: 'text', text: `Created new tab${params.position === 'side' ? ' on the side' : ''}` }] };
+			}
+			case 'close': {
+				const listResult = await vscode.commands.executeCommand<{ tabs: string[] }>('cursor.browserView.listTabs');
+				const tabs = listResult?.tabs ?? [];
+				const targetId = params.index !== undefined ? tabs[params.index] : undefined;
+				await vscode.commands.executeCommand('cursor.browserView.closeTab', targetId);
+				const afterCloseResult = await vscode.commands.executeCommand<{ tabs: string[]; activeTab?: string; lastInteractedTab?: string }>('cursor.browserView.listTabs');
+				const activeViewId = afterCloseResult?.activeTab ?? afterCloseResult?.lastInteractedTab;
+				if (!activeViewId) {
+					return { content: [{ type: 'text', text: 'Tab closed. No remaining tabs.' }] };
+				}
+				return await BrowserTools.snapshot({ viewId: activeViewId });
+			}
+			case 'select': {
+				if (params.index === undefined) {
+					throw new Error('Tab index is required');
+				}
+				const listResult = await vscode.commands.executeCommand<{ tabs: string[] }>('cursor.browserView.listTabs');
+				const tabs = listResult?.tabs ?? [];
+				const targetId = tabs[params.index];
+				if (!targetId) {
+					throw new Error(`Tab ${params.index} not found`);
+				}
+				await vscode.commands.executeCommand('cursor.browserView.selectTab', targetId);
+				return await BrowserTools.snapshot({ viewId: targetId });
+			}
+			default:
+				throw new Error(`Unknown action: ${params.action}`);
+		}
 	}
 };
 
